import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:dcoupon_sdk_flutter_plugin/app_config.dart';
import 'package:dcoupon_sdk_flutter_plugin/constants.dart';
import 'package:dcoupon_sdk_flutter_plugin/dcoupon_listener.dart';
import 'package:dcoupon_sdk_flutter_plugin/enrollmentformview.dart';
import 'package:dcoupon_sdk_flutter_plugin/httprequest.dart';
import 'package:dcoupon_sdk_flutter_plugin/translations.dart';
import 'package:flutter/cupertino.dart';

class Dcoupon {
  BuildContext context;
  bool _isLogged = false;
  String _languageCode;
  String _clientApiToken,
      _clientSecretKey,
      _clientCountry,
      _clientDomain,
      _clientPublisherApiToken,
      _jwt,
      _clientType,
      _urlRedirect,
      _email,
      _externalId;
  Translations translations;

  Dcoupon(BuildContext context) {
    this.context = context;
    _languageCode =
        Locale.cachedLocale != null ? Locale.cachedLocale.languageCode : "en";
    this._loadTranslations(Locale.cachedLocale != null
        ? Locale.cachedLocale
        : new Locale(_languageCode));
  }
  AppConfig appConfig;
  void loadConfig(String contents) {
    try {
      appConfig = AppConfig.loadAsJsonString(contents);
      _clientApiToken = appConfig.clientToken;
      _clientSecretKey = appConfig.clientSecretKey;
      _clientCountry = appConfig.clientCountry;
      _clientDomain = appConfig.clientDomain;
      _clientPublisherApiToken = appConfig.clientPublisherApiToken;
    } catch (ex) {
      print("dcoupon.loadConfig.catchError: " + ex.toString());
    }
  }

  @visibleForTesting
  void setJwt(String jwt) {
    _jwt = jwt;
  }

  @visibleForTesting
  void setClientType(String clientType) {
    _clientType = clientType;
  }

  @visibleForTesting
  void setIsLogged(bool isLogged) {
    _isLogged = isLogged;
  }

  /// Third-party login method.
  ///
  /// @param email      User's email.
  /// @param externalId User's id from your system.
  /// @param alias      User's alias/name.
  /// @param gender     User's gender.
  /// @param birthdate  User's date of birth. Allowed values:'MALE', 'FEMALE', 'OTHER', 'RATHERNOTSAY'. Default 'RATHERNOTSAY'.
  Future<DcouponListener> login(String email, String externalId,
      {String alias, String gender, String birthdate}) async {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.login.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    if (email == null ||
        email.isEmpty ||
        externalId == null ||
        externalId.isEmpty) {
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = translations.text("dcoupon_missing_parameters");
      completer.complete(handleMessage);
      return completer.future;
    }
    _isLogged = false;
    _email = null;
    _externalId = null;
    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        var body = jsonDecode(data.body);
        _jwt = body[HttpInterface.SESSION_TOKEN];
        var headers = jsonDecode(data.headers.toString());
        _clientType = headers[HttpInterface.CLIENT_TYPE];
        _isLogged = true;
        handleMessage.message = Constants.TRUE;
        handleMessage.code = data.statusCode;
      } else if (data.statusCode == HttpInterface.HTTP_MOVED_PERM) {
        handleMessage.message =
            translations.text('dcoupon_mandatory_documents');
        handleMessage.code = data.statusCode;
        var headers = jsonDecode(data.headers.toString());
        _enrollmentForm(headers[HttpInterface.LOCATION]);
      } else {
        handleMessage.message = data.body;
        handleMessage.code = data.statusCode;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    _email = email;
    _externalId = externalId;
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(Constants.CONTENT_TYPE, () => Constants.JSON_UTF_8);
      args.putIfAbsent(Constants.CLIENT_API_TOKEN, () => this._clientApiToken);
      args.putIfAbsent(
          Constants.CLIENT_SECRET_KEY, () => this._clientSecretKey);
      args.putIfAbsent(Constants.CLIENT_COUNTRY, () => this._clientCountry);
      args.putIfAbsent(Constants.EMAIL, () => _email);
      args.putIfAbsent(Constants.EXTERNAL_ID, () => _externalId);
      args.putIfAbsent(Constants.ALIAS, () => alias);
      args.putIfAbsent(Constants.GENDER, () => gender);
      args.putIfAbsent(Constants.BIRTHDATE, () => birthdate);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.THIRD_PARTY_LOGIN);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "login : " + ex.toString());
    }
    return completer.future;
  }

  void _enrollmentForm(String url) {
    _urlRedirect = url;
    var streamController = new StreamController<HttpResponse>();
    var httpRequest = new HttpRequest();
    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        var args = new Map<dynamic, dynamic>();
        args.putIfAbsent(HttpInterface.JSON_RESPONSE, () => data.body);
        args.putIfAbsent(Constants.EXTERNAL_ID, () => _externalId);
        args.putIfAbsent(Constants.CLIENT_API_TOKEN, () => _clientApiToken);
        args.putIfAbsent(
            Constants.JWT,
            () => _urlRedirect
                .substring(_urlRedirect.indexOf(Constants.JWT + "=") + 4));
        Navigator.of(context).push(
          PageRouteBuilder(
            opaque: false,
            pageBuilder: (BuildContext context, _, __) {
              return EnrollmentFormView(
                  args: args, handlerAcceptEnrollment: _acceptEnrollment);
            },
          ),
        );
      }
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(Constants.CLIENT_API_TOKEN, () => this._clientApiToken);
      args.putIfAbsent(Constants.JWT,
          () => url.substring(url.indexOf(Constants.JWT + "=") + 4));
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.THIRD_PARTY_ENROLLMENT_FORM);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "_enrollmentForm : " + ex.toString());
    }
  }

  void _acceptEnrollment(Map<dynamic, dynamic> map) {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    streamController.stream.listen((data) {
      var handleMessage = new DcouponListener();
      if (data.statusCode == HttpInterface.HTTP_OK) {
        if (data.statusCode == HttpInterface.HTTP_OK) {
          var body = jsonDecode(data.body);
          _jwt = body[HttpInterface.SESSION_TOKEN];
          var headers = jsonDecode(data.headers.toString());
          _clientType = headers[HttpInterface.CLIENT_TYPE];
          _isLogged = true;
          handleMessage.message = Constants.TRUE;
          handleMessage.code = data.statusCode;
          completer.complete(handleMessage);
        }
      } else {
        handleMessage.message = Constants.ERROR;
        handleMessage.code = data.statusCode;
        completer.complete(handleMessage);
      }
      Navigator.pop(context);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(
          Constants.JWT,
          () => _urlRedirect
              .substring(_urlRedirect.indexOf(Constants.JWT + "=") + 4));
      args.putIfAbsent(Constants.CLIENT_API_TOKEN, () => _clientApiToken);
      args.putIfAbsent(Constants.CLIENT_SECRET_KEY, () => _clientSecretKey);
      args.putIfAbsent(Constants.CLIENT_DOMAIN, () => _clientDomain);
      args.putIfAbsent(Constants.EXTERNAL_ID, () => _externalId);
      args.putIfAbsent(Constants.EMAIL, () => _email);
      args.putIfAbsent(Constants.GENDER, () => map[Constants.GENDER]);
      args.putIfAbsent(Constants.BIRTHDATE, () => map[Constants.BIRTHDATE]);
      args.putIfAbsent(Constants.ALIAS, () => map[Constants.ALIAS]);
      var arrayList = new List<Map<dynamic, dynamic>>();
      List<dynamic> jsonArray = map[Constants.LEGAL_DOCUMENTS_ACCEPTED];
      for (int i = 0; i < jsonArray.length; i++) {
        var jsonObject = jsonArray[i];
        var legalDocsAccepted = new Map<dynamic, dynamic>();
        legalDocsAccepted.putIfAbsent(
            Constants.DOC_TYPE, () => jsonObject[Constants.DOC_TYPE]);
        legalDocsAccepted.putIfAbsent(
            Constants.DOC_UUID, () => jsonObject[Constants.DOC_UUID]);
        arrayList.add(legalDocsAccepted);
      }
      args.putIfAbsent(Constants.LEGAL_DOCUMENTS_ACCEPTED, () => arrayList);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.THIRD_PARTY_ACCEPT_ENROLLMENT);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "_acceptEnrollment : " + ex.toString());
    }
  }

  /// Get a list of coupons.
  ///
  /// @param start  Index of first store to be returned. 0 by default.
  /// @param limit  Maximum number of items to return. 10 by default.
  /// @param sortBy How the result set is sorted. Allowed values: 'NEWEST', 'ENDING', 'VALUE', 'DEFAULT'. Default 'DEFAULT'.
  Future<DcouponListener> getCoupons(
      {int start, int limit, String sortBy}) async {
    var completer = new Completer<DcouponListener>();
    var streamController = new StreamController<HttpResponse>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.getCoupons.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    if (!isLogged()) {
      handleMessage.code = HttpInterface.HTTP_UNAUTHORIZED;
      handleMessage.message = translations.text("dcoupon_is_logged");
      completer.complete(handleMessage);
      return completer.future;
    }
    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        handleMessage.code = data.statusCode;
        Map<dynamic, dynamic> body = jsonDecode(data.body);
        Map<dynamic, dynamic> couponsList = body[Constants.COUPON_DETAIL_LIST];
        couponsList.remove(HttpInterface.RESPONSE);
        handleMessage.message = jsonEncode(couponsList);
      } else {
        handleMessage.code = data.statusCode;
        handleMessage.message = Constants.ERROR;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(
          HttpInterface.AUTHORIZATION, () => HttpInterface.BEARER_ + _jwt);
      args.putIfAbsent(Constants.START, () => start);
      args.putIfAbsent(Constants.LIMIT, () => limit);
      args.putIfAbsent(Constants.SORT_BY, () => sortBy);
      if (_clientType == Constants.CLIENT_TYPE_RETAILER) {
        args.putIfAbsent(Constants.RETAILER_TOKENS, () => this._clientApiToken);
      } else if (_clientType == Constants.CLIENT_TYPE_COMPANY) {
        args.putIfAbsent(Constants.COMPANY_TOKENS, () => this._clientApiToken);
      }
      args.putIfAbsent(Constants.LANG, () => _languageCode);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.USER_COUPONS);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "getCoupons : " + ex.toString());
    }
    return completer.future;
  }

  /// Get the details of one coupon.
  ///
  /// @param couponId The identifier of the user's coupon.
  Future<DcouponListener> getCouponDetail(int couponId) async {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();
    if (!_isSetUp()) {
      print("dcoupon.getCouponDetail.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    if (!isLogged()) {
      handleMessage.code = HttpInterface.HTTP_UNAUTHORIZED;
      handleMessage.message = translations.text("dcoupon_is_logged");
      completer.complete(handleMessage);
      return completer.future;
    }
    if (couponId == 0) {
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = translations.text("dcoupon_missing_parameters");
      completer.complete(handleMessage);
      return completer.future;
    }
    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        var body = jsonDecode(data.body);
        if (body[Constants.GENERIC_RESPONSE][HttpInterface.RESPONSE_CODE] ==
            111) {
          handleMessage.code = HttpInterface.HTTP_NOT_FOUND;
          handleMessage.message =
              body[Constants.GENERIC_RESPONSE][Constants.RESPONSE_MSG];
        } else if (body[Constants.GENERIC_RESPONSE]
                [HttpInterface.RESPONSE_CODE] ==
            0) {
          handleMessage.code = data.statusCode;
          handleMessage.message = data.body;
        } else {
          handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
          handleMessage.message = data.body;
        }
      } else {
        handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
        handleMessage.message = Constants.ERROR;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(
          HttpInterface.AUTHORIZATION, () => HttpInterface.BEARER_ + _jwt);
      args.putIfAbsent(Constants.LANG, () => _languageCode);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.USER_COUPON);
      httpRequest.setPathParameter(couponId);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "getCouponDetail : " + ex.toString());
    }
    return completer.future;
  }

  /// Set the activated status on a coupon.
  ///
  /// @param couponId The identifier of the user's coupon.
  Future<DcouponListener> activateCoupon(int couponId) async {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.activateCoupon.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    if (!isLogged()) {
      handleMessage.code = HttpInterface.HTTP_UNAUTHORIZED;
      handleMessage.message = translations.text("dcoupon_is_logged");
      completer.complete(handleMessage);
      return completer.future;
    }
    if (couponId == 0) {
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = translations.text("dcoupon_missing_parameters");
      completer.complete(handleMessage);
      return completer.future;
    }
    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        var body = jsonDecode(data.body);
        try {
          if (body[HttpInterface.RESPONSE_CODE] == 111) {
            handleMessage.code = HttpInterface.HTTP_NOT_FOUND;
            handleMessage.message = body[Constants.RESPONSE_MSG];
          } else if (body[HttpInterface.RESPONSE_CODE] == 0) {
            handleMessage.code = data.statusCode;
            handleMessage.message = data.body;
          } else {
            handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
            handleMessage.message = data.body;
          }
        } catch (ex) {
          print(ex.toString());
          handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
          handleMessage.message = Constants.ERROR;
        }
      } else {
        handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
        handleMessage.message = Constants.ERROR;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(
          HttpInterface.AUTHORIZATION, () => HttpInterface.BEARER_ + _jwt);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.USER_COUPON_ACTIVATE);
      httpRequest.setPathParameter(couponId);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "activateCoupon : " + ex.toString());
    }
    return completer.future;
  }

  /// Set the deactivated status on a coupon.
  ///
  /// @param couponId The identifier of the user's coupon.
  Future<DcouponListener> deactivateCoupon(int couponId) async {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.deactivateCoupon.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    if (!isLogged()) {
      handleMessage.code = HttpInterface.HTTP_UNAUTHORIZED;
      handleMessage.message = translations.text("dcoupon_is_logged");
      completer.complete(handleMessage);
      return completer.future;
    }
    if (couponId == 0) {
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = translations.text("dcoupon_missing_parameters");
      completer.complete(handleMessage);
      return completer.future;
    }
    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        try {
          var body = jsonDecode(data.body);
          if (body[HttpInterface.RESPONSE_CODE] == 111) {
            handleMessage.code = HttpInterface.HTTP_NOT_FOUND;
            handleMessage.message = body[Constants.RESPONSE_MSG];
          } else if (body[HttpInterface.RESPONSE_CODE] == 0) {
            handleMessage.code = data.statusCode;
            handleMessage.message = data.body;
          } else {
            handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
            handleMessage.message = data.body;
          }
        } catch (ex) {
          print(ex.toString());
          handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
          handleMessage.message = Constants.ERROR;
        }
      } else {
        handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
        handleMessage.message = Constants.ERROR;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(
          HttpInterface.AUTHORIZATION, () => HttpInterface.BEARER_ + _jwt);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.USER_COUPON_DEACTIVATE);
      httpRequest.setPathParameter(couponId);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "deactivateCoupon : " + ex.toString());
    }
    return completer.future;
  }

  /// Get an available redemption code.
  Future<DcouponListener> getRedemptionCode() async {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.deactivateCoupon.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    if (!isLogged()) {
      handleMessage.code = HttpInterface.HTTP_UNAUTHORIZED;
      handleMessage.message = translations.text("dcoupon_is_logged");
      completer.complete(handleMessage);
      return completer.future;
    }
    streamController.stream.listen((data) {
      try {
        if (data.statusCode == HttpInterface.HTTP_OK) {
          Map<dynamic, dynamic> body = jsonDecode(data.body);
          body.remove(Constants.GENERIC_RESPONSE);
          handleMessage.code = data.statusCode;
          handleMessage.message = jsonEncode(body);
        } else {
          handleMessage.code = data.statusCode;
          handleMessage.message = Constants.ERROR;
        }
      } catch (ex) {
        handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
        handleMessage.message = Constants.ERROR;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(
          HttpInterface.AUTHORIZATION, () => HttpInterface.BEARER_ + _jwt);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.USER_REDEMPTION_CODE);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "getRedemptionCode : " + ex.toString());
    }
    return completer.future;
  }

  /// Create multiple coupons by promotion tokens.
  ///
  /// @param promotionTokens Promotions tokens that identifies the promotions.
  /// @param latitude        User latitude coordinate.
  /// @param longitude       User longitude coordinate.
  /// @param source          Source from the request.
  /// @param transId         Transaction identifier if the promotion requires it.
  /// @param crmId           The customer relationship management identifier.
  Future<DcouponListener> createCoupons(List<String> promotionTokens,
      {String latitude,
      String longitude,
      String source,
      String transId,
      String crmId}) async {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.createCoupons.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    if (!isLogged()) {
      handleMessage.code = HttpInterface.HTTP_UNAUTHORIZED;
      handleMessage.message = translations.text("dcoupon_is_logged");
      completer.complete(handleMessage);
      return completer.future;
    }
    if (promotionTokens == null || promotionTokens.isEmpty) {
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = translations.text("dcoupon_missing_parameters");
      completer.complete(handleMessage);
      return completer.future;
    }
    streamController.stream.listen((data) {
      try {
        if (data.statusCode == HttpInterface.HTTP_OK) {
          Map<dynamic, dynamic> body = jsonDecode(data.body);
          body.remove(Constants.GENERIC_RESPONSE);
          handleMessage.code = data.statusCode;
          handleMessage.message = jsonEncode(body);
        } else {
          handleMessage.code = data.statusCode;
          handleMessage.message = Constants.ERROR;
        }
      } catch (ex) {
        handleMessage.code = HttpInterface.HTTP_INTERNAL_ERROR;
        handleMessage.message = Constants.ERROR;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(
          HttpInterface.AUTHORIZATION, () => HttpInterface.BEARER_ + _jwt);
      args.putIfAbsent(Constants.CONTENT_TYPE, () => Constants.JSON_UTF_8);
      args.putIfAbsent(
          Constants.API_TOKEN_PUBLISHER, () => this._clientPublisherApiToken);
      var metaCouponTokens = new Map<dynamic, List<String>>();
      metaCouponTokens.putIfAbsent(
          Constants.API_TOKEN_META_COUPON, () => promotionTokens);
      args.putIfAbsent(Constants.META_COUPON_TOKENS, () => metaCouponTokens);
      var ucouponType = new Map<dynamic, dynamic>();
      ucouponType.putIfAbsent(Constants.CREATION_LATITUDE, () => latitude);
      ucouponType.putIfAbsent(Constants.CREATION_LONGITUDE, () => longitude);
      ucouponType.putIfAbsent(
          Constants.SOURCE,
          () =>
              Constants.SDK +
              "_" +
              _clientType +
              "#" +
              _clientApiToken +
              "#" +
              (source != null ? source : ""));
      ucouponType.putIfAbsent(Constants.TRANS_ID, () => transId);
      ucouponType.putIfAbsent(Constants.CRM_ID, () => crmId);
      args.putIfAbsent(Constants.UCUPON_TYPE, () => ucouponType);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.USER_CREATE_COUPONS);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "createCoupons : " + ex.toString());
    }
    return completer.future;
  }

  /// Get all companies, categories and retailers with active promotions associated to the publisher api token.
  ///
  /// @param start Index of first store to be returned. 0 by default.
  /// @param limit Maximum number of items to return. 10 by default.
  Future<DcouponListener> getFilters({int start, int limit}) async {
    var streamControllerCategories = new StreamController<HttpResponse>();
    var streamControllerCompanies = new StreamController<HttpResponse>();
    var streamControllerRetailers = new StreamController<HttpResponse>();
    var httpRequestCategories = new HttpRequest();
    var httpRequestCompanies = new HttpRequest();
    var httpRequestRetailers = new HttpRequest();
    var completer = new Completer<DcouponListener>();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.getFilters.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    int i = 0;
    var jsonToReturn = new Map<dynamic, dynamic>();
    streamControllerCategories.stream.listen((data) {
      i++;
      if (data.statusCode == HttpInterface.HTTP_OK)
        jsonToReturn.putIfAbsent(
            Constants.FILTERS_CATEGORIES, () => jsonDecode(data.body));
      if (i == 3) {
        handleMessage.message = jsonEncode(jsonToReturn);
        handleMessage.code = data.statusCode;
        completer.complete(handleMessage);
      }
      streamControllerCategories.close();
    });
    streamControllerCompanies.stream.listen((data) {
      i++;
      if (data.statusCode == HttpInterface.HTTP_OK)
        jsonToReturn.putIfAbsent(
            Constants.FILTERS_COMPANIES, () => jsonDecode(data.body));
      if (i == 3) {
        handleMessage.message = jsonEncode(jsonToReturn);
        handleMessage.code = data.statusCode;
        completer.complete(handleMessage);
      }
      streamControllerCompanies.close();
    });
    streamControllerRetailers.stream.listen((data) {
      i++;
      if (data.statusCode == HttpInterface.HTTP_OK)
        jsonToReturn.putIfAbsent(
            Constants.FILTERS_RETAILERS, () => jsonDecode(data.body));
      if (i == 3) {
        handleMessage.message = jsonEncode(jsonToReturn);
        handleMessage.code = data.statusCode;
        completer.complete(handleMessage);
      }
      streamControllerRetailers.close();
    });

    var args = new Map<dynamic, dynamic>();
    args.putIfAbsent(
        Constants.API_TOKEN_PUBLISHER, () => this._clientPublisherApiToken);
    args.putIfAbsent(Constants.CLIENT_API_TOKEN, () => _clientApiToken);
    args.putIfAbsent(Constants.CLIENT_SECRET_KEY, () => _clientSecretKey);
    args.putIfAbsent(Constants.PUBLISHER_ID, () => _clientPublisherApiToken);
    args.putIfAbsent(Constants.LANG, () => _languageCode);
    args.putIfAbsent(Constants.COUNTRY, () => _clientCountry);
    args.putIfAbsent(Constants.START, () => start);
    args.putIfAbsent(Constants.LIMIT, () => limit);

    httpRequestCategories.setStreamController(streamControllerCategories);
    try {
      httpRequestCategories.setArguments(args);
      httpRequestCategories.setDomain(this._clientDomain);
      httpRequestCategories.setMethod(Constants.PUBLISHER_CATEGORIES);
      httpRequestCategories.execute();
    } on Exception catch (ex) {
      throwException(streamControllerCategories,
          "getFiltersCategories : " + ex.toString());
    }
    httpRequestCompanies.setStreamController(streamControllerCompanies);
    try {
      httpRequestCompanies.setArguments(args);
      httpRequestCompanies.setDomain(this._clientDomain);
      httpRequestCompanies.setMethod(Constants.PUBLISHER_COMPANIES);
      httpRequestCompanies.execute();
    } on Exception catch (ex) {
      throwException(
          streamControllerCompanies, "getFiltersCompanies : " + ex.toString());
    }
    httpRequestRetailers.setStreamController(streamControllerRetailers);
    try {
      httpRequestRetailers.setArguments(args);
      httpRequestRetailers.setDomain(this._clientDomain);
      httpRequestRetailers.setMethod(Constants.PUBLISHER_RETAILERS);
      httpRequestRetailers.execute();
    } on Exception catch (ex) {
      throwException(
          streamControllerRetailers, "getFiltersRetailers : " + ex.toString());
    }
    return completer.future;
  }

  /// Get all active promotions associated to the publisher api token.
  ///
  /// @param start          Index of first store to be returned. 0 by default.
  /// @param limit          Maximum number of items to return. 10 by default.
  /// @param categoryIds    Filter offers from this/these category/ies.
  /// @param storeIds       Filter offers redeemable in those stores form the selected retailer.
  /// @param retailerTokens Filter offers redeemable in this/these retailer/s.
  /// @param companyTokens  Filter offers by this/these company/s.
  /// @param textSearch     Text to search for into promotion name, promotion description and company owner of the promotions.
  /// @param zipcode        Zip code to filter the search.
  /// @param latitude       User latitude coordinate.
  /// @param longitude      User longitude coordinate.
  /// @param radius         Radius to make the search for coupons available.
  /// @param orderBy        How the result set is sorted.
  Future<DcouponListener> getPromotions(
      {int start,
      int limit,
      List<String> categoryIds,
      List<String> storeIds,
      List<String> retailerTokens,
      List<String> companyTokens,
      String textSearch,
      String zipcode,
      String latitude,
      String longitude,
      String radius,
      String orderBy}) async {
    var streamController = new StreamController<HttpResponse>();
    var httpRequest = new HttpRequest();
    var completer = new Completer<DcouponListener>();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.getPromotions.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }

    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        handleMessage.message = data.body;
        handleMessage.code = data.statusCode;
      } else {
        handleMessage.message = data.body;
        handleMessage.code = data.statusCode;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(Constants.CLIENT_API_TOKEN, () => _clientApiToken);
      args.putIfAbsent(Constants.CLIENT_SECRET_KEY, () => _clientSecretKey);
      args.putIfAbsent(Constants.CLIENT_DOMAIN, () => _clientDomain);
      args.putIfAbsent(Constants.PUBLISHER_ID, () => _clientPublisherApiToken);
      args.putIfAbsent(Constants.LANG, () => _languageCode);
      args.putIfAbsent(Constants.COUNTRY, () => _clientCountry);
      args.putIfAbsent(Constants.START, () => start);
      args.putIfAbsent(Constants.LIMIT, () => limit);
      args.putIfAbsent(Constants.TEXT_SEARCH, () => textSearch);
      args.putIfAbsent(Constants.ZIPCODE, () => zipcode);
      args.putIfAbsent(Constants.LATITUDE, () => latitude);
      args.putIfAbsent(Constants.LONGITUDE, () => longitude);
      args.putIfAbsent(Constants.RADIUS, () => radius);
      args.putIfAbsent(Constants.ORDER_BY, () => orderBy);
      if (categoryIds != null && categoryIds.length > 0) {
        args.putIfAbsent(Constants.CATEGORY_IDS,
            () => categoryIds.join(Constants.SEPARATOR));
      }
      if (storeIds != null && storeIds.length > 0) {
        args.putIfAbsent(
            Constants.STORE_IDS, () => storeIds.join(Constants.SEPARATOR));
      }
      switch (_clientType) {
        case Constants.CLIENT_TYPE_COMPANY:
          args.putIfAbsent(Constants.COMPANY_TOKENS, () => _clientApiToken);
          if (retailerTokens != null && retailerTokens.length > 0) {
            args.putIfAbsent(Constants.RETAILER_TOKENS,
                () => retailerTokens.join(Constants.SEPARATOR));
          }
          break;
        case Constants.CLIENT_TYPE_RETAILER:
          args.putIfAbsent(Constants.RETAILER_TOKENS, () => _clientApiToken);
          if (companyTokens != null && companyTokens.length > 0) {
            args.putIfAbsent(Constants.COMPANY_TOKENS,
                () => companyTokens.join(Constants.SEPARATOR));
          }
          break;
        case Constants.CLIENT_TYPE_PUBLISHER:
          if (retailerTokens != null && retailerTokens.length > 0) {
            args.putIfAbsent(Constants.RETAILER_TOKENS,
                () => retailerTokens.join(Constants.SEPARATOR));
          }
          if (companyTokens != null && companyTokens.length > 0) {
            args.putIfAbsent(Constants.COMPANY_TOKENS,
                () => companyTokens.join(Constants.SEPARATOR));
          }
          break;
      }
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.PUBLISHER_PROMOTIONS);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "getPromotions : " + ex.toString());
    }
    return completer.future;
  }

  /// Get the details of a promotion.
  ///
  /// @param promotionToken The token that identifies this promotion.
  Future<DcouponListener> getPromotionDetail(String promotionToken) async {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.getPromotionDetail.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    if (promotionToken == null || promotionToken.isEmpty) {
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = translations.text("dcoupon_missing_parameters");
      completer.complete(handleMessage);
      return completer.future;
    }

    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        handleMessage.message = data.body;
        handleMessage.code = data.statusCode;
      } else {
        handleMessage.message = data.body;
        handleMessage.code = data.statusCode;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(Constants.CLIENT_API_TOKEN, () => _clientApiToken);
      args.putIfAbsent(Constants.CLIENT_SECRET_KEY, () => _clientSecretKey);
      args.putIfAbsent(Constants.CLIENT_DOMAIN, () => _clientDomain);
      args.putIfAbsent(Constants.PUBLISHER_ID, () => _clientPublisherApiToken);
      args.putIfAbsent(Constants.LANG, () => _languageCode);
      args.putIfAbsent(Constants.COUNTRY, () => _clientCountry);
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.PUBLISHER_PROMOTION);
      httpRequest.setPathParameter(promotionToken);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "getPromotionDetail : " + ex.toString());
    }
    return completer.future;
  }

  /// Get a list of stores associated with a promotion token identified.
  ///
  /// @param start          Index of first store to be returned. 0 by default.
  /// @param limit          Maximum number of items to return. 10 by default.
  /// @param promotionToken The token that identifies the promotion.
  /// @param retailerTokens Token from the retailer.
  /// @param zipcode        Zip code to filter the search.
  /// @param latitude       User latitude coordinate.
  /// @param longitude      User longitude coordinate.
  /// @param radius         Radius to make the search for coupons available.
  Future<DcouponListener> getStores(
      {int start,
      int limit,
      String promotionToken,
      List<String> retailerTokens,
      String zipcode,
      String latitude,
      String longitude,
      String radius}) async {
    var streamController = new StreamController<HttpResponse>();
    var completer = new Completer<DcouponListener>();
    var httpRequest = new HttpRequest();
    var handleMessage = new DcouponListener();

    if (!_isSetUp()) {
      print("dcoupon.getStores.isSetUp.error");
      handleMessage.code = HttpInterface.HTTP_BAD_REQUEST;
      handleMessage.message = "Please, configure the dcoupon sdk correctly";
      completer.complete(handleMessage);
      return completer.future;
    }
    streamController.stream.listen((data) {
      if (data.statusCode == HttpInterface.HTTP_OK) {
        handleMessage.message = data.body;
        handleMessage.code = data.statusCode;
      } else {
        handleMessage.message = data.body;
        handleMessage.code = data.statusCode;
      }
      completer.complete(handleMessage);
      streamController.close();
    });
    httpRequest.setStreamController(streamController);
    try {
      var args = new Map<dynamic, dynamic>();
      args.putIfAbsent(Constants.CLIENT_API_TOKEN, () => _clientApiToken);
      args.putIfAbsent(Constants.CLIENT_SECRET_KEY, () => _clientSecretKey);
      args.putIfAbsent(Constants.PUBLISHER_ID, () => _clientPublisherApiToken);
      args.putIfAbsent(Constants.LANG, () => _languageCode);
      args.putIfAbsent(Constants.COUNTRY, () => _clientCountry);
      args.putIfAbsent(Constants.START, () => start);
      args.putIfAbsent(Constants.LIMIT, () => limit);
      args.putIfAbsent(Constants.PROMOTION_TOKEN, () => promotionToken);
      args.putIfAbsent(Constants.ZIPCODE, () => zipcode);
      args.putIfAbsent(Constants.LATITUDE, () => latitude);
      args.putIfAbsent(Constants.LONGITUDE, () => longitude);
      args.putIfAbsent(Constants.RADIUS, () => radius);
      switch (_clientType) {
        case Constants.CLIENT_TYPE_RETAILER:
          args.putIfAbsent(Constants.RETAILER_TOKENS, () => _clientApiToken);
          break;
        case Constants.CLIENT_TYPE_COMPANY:
        case Constants.CLIENT_TYPE_PUBLISHER:
          if (retailerTokens != null && retailerTokens.length > 0) {
            args.putIfAbsent(Constants.RETAILER_TOKENS,
                () => retailerTokens.join(Constants.SEPARATOR));
          }
          break;
      }
      httpRequest.setArguments(args);
      httpRequest.setDomain(this._clientDomain);
      httpRequest.setMethod(Constants.PUBLISHER_STORES);
      httpRequest.execute();
    } on Exception catch (ex) {
      throwException(streamController, "getStores : " + ex.toString());
    }
    return completer.future;
  }

  /// @throws NotImplementedException function not implemented
  Future<DcouponListener> getRedemptionHistory() async {
    throw UnimplementedError(
        "getRedemptionHistory : " + Constants.NOT_IMPLEMENTED);
  }

  /// @throws NotImplementedException function not implemented
  Future<DcouponListener> addLoyaltyCard() async {
    throw UnimplementedError("addLoyaltyCard : " + Constants.NOT_IMPLEMENTED);
  }

  bool isLogged() {
    return _isLogged;
  }

  bool _isSetUp() {
    var message = new List<String>();
    if (this._clientApiToken == null || this._clientApiToken.isEmpty) {
      message.add("com.dcoupon.sdk.ClientToken");
    }
    if (this._clientSecretKey == null || this._clientSecretKey.isEmpty) {
      message.add("com.dcoupon.sdk.ClientSecretKey");
    }
    if (this._clientCountry == null || this._clientCountry.isEmpty) {
      message.add("com.dcoupon.sdk.ClientCountry");
    }
    if (this._clientDomain == null || this._clientDomain.isEmpty) {
      message.add("com.dcoupon.sdk.ClientDomain");
    } else {
      if (!(this._clientDomain == "local" ||
          this._clientDomain == "test" ||
          this._clientDomain == "dev" ||
          this._clientDomain == "dcoupon.eu" ||
          this._clientDomain == "dcoupon.com"))
        message.add("com.dcoupon.sdk.ClientDomain not allowed");
    }
    if (this._clientPublisherApiToken == null ||
        this._clientPublisherApiToken.isEmpty) {
      message.add("com.dcoupon.sdk.ClientPublisherApiToken");
    }
    if (message.isNotEmpty) {
      print("Missed manadatory parameter/s:");
      message.forEach((str) {
        print(str);
      });
      return false;
    } else {
      return true;
    }
  }

  void _loadTranslations(Locale locale) async {
    try {
      //print("dcoupon.translations.loading...");
      translations = await Translations.load(locale);
      //print("translations.loaded");
    } catch (ex) {
      print("dcoupon.dcoupon.loadTranslations.catchError: " + ex.toString());
    }
  }

  void throwException(StreamController streamController, String message) {
    HttpResponse httpResponse = new HttpResponse();
    httpResponse.statusCode = HttpInterface.HTTP_INTERNAL_ERROR;
    httpResponse.headers = new Map();
    httpResponse.body = message;
    streamController.sink.add(httpResponse);
  }
}
